use nom::branch::alt;
use nom::bytes::complete::tag;
use nom::character::complete::digit1;
use nom::character::complete::one_of;
use nom::character::complete::space0;
use nom::combinator::eof;
use nom::combinator::not;
use nom::combinator::recognize;
use nom::combinator::verify;
use nom::multi::many0;
use nom::sequence::tuple;

use crate::parser::element::element;
use crate::parser::parser_context::ChainBehavior;
use crate::parser::parser_context::ContextElement;
use crate::parser::parser_context::ExitMatcherNode;
use crate::parser::util::get_consumed;
use crate::parser::util::start_of_line;

use super::error::CustomError;
use super::error::MyError;
use super::error::Res;
use super::greater_element::PlainList;
use super::greater_element::PlainListItem;
use super::parser_with_context::parser_with_context;
use super::util::non_whitespace_character;
use super::Context;

#[allow(dead_code)]
pub fn plain_list<'r, 's>(context: Context<'r, 's>, input: &'s str) -> Res<&'s str, PlainList<'s>> {
    todo!()
}

#[allow(dead_code)]
pub fn plain_list_item<'r, 's>(
    context: Context<'r, 's>,
    input: &'s str,
) -> Res<&'s str, PlainListItem<'s>> {
    not(|i| context.check_exit_matcher(i))(input)?;
    start_of_line(context, input)?;
    let (remaining, leading_whitespace) = space0(input)?;
    // It is fine that we get the indent level using the number of bytes rather than the number of characters because nom's space0 only matches space and tab (0x20 and 0x09)
    let indent_level = leading_whitespace.len();
    let parser_context = context
        .with_additional_node(ContextElement::ExitMatcherNode(ExitMatcherNode {
            exit_matcher: ChainBehavior::AndParent(Some(&plain_list_item_end)),
        }))
        .with_additional_node(ContextElement::ListItem(indent_level));

    let element_matcher = parser_with_context!(element)(&parser_context);
    let (remaining, (bull, _ws)) = tuple((bullet, space0))(remaining)?;
    let (remaining, contents) = many0(element_matcher)(remaining)?;
    let source = get_consumed(input, remaining);

    Ok((
        remaining,
        PlainListItem {
            source,
            bullet: bull,
            contents,
        },
    ))
}

fn bullet<'s>(i: &'s str) -> Res<&'s str, &'s str> {
    alt((
        tag("*"),
        tag("-"),
        tag("+"),
        recognize(tuple((counter, alt((tag("."), tag(")")))))),
    ))(i)
}

fn counter<'s>(i: &'s str) -> Res<&'s str, &'s str> {
    alt((recognize(one_of("abcdefghijklmnopqrstuvwxyz")), digit1))(i)
}

fn plain_list_item_end<'r, 's>(context: Context<'r, 's>, input: &'s str) -> Res<&'s str, &'s str> {
    let plain_list_item_matcher = parser_with_context!(plain_list_item)(context);
    let line_indented_lte_matcher = parser_with_context!(line_indented_lte)(context);
    alt((
        recognize(plain_list_item_matcher),
        line_indented_lte_matcher,
        eof,
    ))(input)
}

fn line_indented_lte<'r, 's>(context: Context<'r, 's>, input: &'s str) -> Res<&'s str, &'s str> {
    let current_item_indent_level: &usize =
        get_context_item_indent(context).ok_or(nom::Err::Error(CustomError::MyError(MyError(
            "Not inside a plain list item",
        ))))?;

    start_of_line(context, input)?;

    let matched = recognize(verify(
        tuple((space0::<&str, _>, non_whitespace_character)),
        // It is fine that we get the indent level using the number of bytes rather than the number of characters because nom's space0 only matches space and tab (0x20 and 0x09)
        |(_space0, _anychar)| _space0.len() <= *current_item_indent_level,
    ))(input)?;

    Ok(matched)
}

fn get_context_item_indent<'r, 's>(context: Context<'r, 's>) -> Option<&'r usize> {
    for thing in context.iter() {
        match thing.get_data() {
            ContextElement::ListItem(depth) => return Some(depth),
            _ => {}
        };
    }
    None
}

#[cfg(test)]
mod tests {
    use crate::parser::parser_context::ContextElement;
    use crate::parser::parser_context::ContextTree;
    use crate::parser::parser_with_context::parser_with_context;

    use super::*;

    #[test]
    fn plain_list_item_empty() {
        let input = "1.";
        let initial_context: ContextTree<'_, '_> = ContextTree::new();
        let document_context =
            initial_context.with_additional_node(ContextElement::DocumentRoot(input));
        let plain_list_item_matcher = parser_with_context!(plain_list_item)(&document_context);
        let (remaining, result) = plain_list_item_matcher(input).unwrap();
        assert_eq!(remaining, "");
        assert_eq!(result.source, "1.");
    }

    #[test]
    fn plain_list_item_simple() {
        let input = "1. foo";
        let initial_context: ContextTree<'_, '_> = ContextTree::new();
        let document_context =
            initial_context.with_additional_node(ContextElement::DocumentRoot(input));
        let plain_list_item_matcher = parser_with_context!(plain_list_item)(&document_context);
        let (remaining, result) = plain_list_item_matcher(input).unwrap();
        assert_eq!(remaining, "");
        assert_eq!(result.source, "1. foo");
    }
}
